#!/usr/bin/perl -w
#  
#  Copyright (c) 2002 Steve Slaven, All Rights Reserved.
#  
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston,
#  MA 02111-1307 USA
#  

# This takes a message on STDIN and outputs the message to propogate on
# STDOUT.  It will scan and sanitize any attachments, and if any problems
# happen, it will remove the offending attachments and attach the virus
# log, and pass thru the modified message and notify the administrator
use strict;
use vars qw( %ENV %opts );

use MIME::Parser;
use MIME::Entity;
use Mail::Sender;
use Getopt::Std;
use FindBin;
use Archive::Zip;

my $VER = '0.88';
delete( $ENV{ PATH } ) if exists( $ENV{ PATH } );

### CONFIGURE
getopts( 'p:a:s:lm:qhS', \%opts );

die( qq{
scanmail.pl v$VER
Scans incoming email for viruses using f-prot
Author: Steve Slaven - http://hoopajoo.net

Usage: $0 [-lhqS] [-p LOADAVG] [-a ADMIN ADDRESS] [-s SCANPATH]
	[-f FPROTPATH] [-m MESSAGEFILE]

	-h	This help
	-l	Always attach virus scan log to all messages
	-q	Quarantine infected messages instead of deleting them
	-p	Don't scan mail if load average for last 1 min higher
		than LOADAVG
	-a	Admin email address, defaults to root\@localhost
	-s	Scanning temp dir, defaults to /var/spool/virus
	-f	Path to f-prot, defaults to /usr/local/bin/f-prot
	-S	Silent admin ops, do not email the admin all virus
		problems
	-m	Message file to use when an infected message is received,
		if not specified uses a default internal message

Meant to be used from procmail, acts as a filter reading on STDIN
and printing the resulting message on STDOUT.  An example procmailrc
would contain:

:0 f
| /usr/local/vscan/scanmail.pl

With any options to use with scanmail.pl

F-Prot can be downloaded from ftp://ftp.f-prot.com

} ) if $opts{ h };

# Set this to the path to send stuff
my $always_attach_log = $opts{ l } || 0; # Set to 1 to attach the log from every scan
my $scan_path = $opts{ s } || "/var/spool/virus";
my $administrator = $opts{ a } || 'Virus Alert <root@localhost>';
my $fprot = $opts{ f } || '/usr/local/bin/f-prot';
my $quarantine = $opts{ q } || 0; # Set to 1 to enable quarantine

####################

# Check loadavg
if( $opts{ p } && -f '/proc/loadavg')
{
	my $last_min;
	open( LOAD, "/proc/loadavg" );
	( $last_min ) = <LOAD> =~ /^(\S+)/;
	close( LOAD );

	if( $last_min > $opts{ p } )
	{
		logmsg( "Loadavg too high, not scanning" );
		print $_ while <STDIN>;
		exit;
	}
}

if( ! -d $scan_path )
{
	# Pass thru message, and notify sysadmin that we are not scanning
	print $_ while <STDIN>;

	mail_admin( 'WARNING: COULD NOT LOCATE SCANDIR',
		"Scandir '$scan_path' does not exist, scanning disabled" );

	exit;
}

if( ! -d "$scan_path/quarantine" )
{
	logmsg( "Creating quarantine directory" );
	mkdir "$scan_path/quarantine", 0755;
}

logmsg( "Mail scanner started at " . scalar( localtime() ) );
my $first_line = <STDIN>;
my $virus_found = 0;

my $parser = new MIME::Parser;
$parser -> output_under( $scan_path );

my $en = $parser -> parse( \*STDIN ) || die( "Could not parse: $?" );

# Grab some atrtibutes
my $from = $en -> head() -> get( "from" );
my $subject = $en -> head() -> get( "subject" );
my $to = $en -> head() -> get( "to" );
my $edump = "\n";

chomp( $from );
chomp( $subject );
chomp( $to );

for( keys( %ENV ) )
{
	$edump .= "$_: " . $ENV{ $_ } . "\n";
}

# Wait for scanning to complete
my $tmpdir = $parser -> filer() -> output_dir( $en -> head() );

# Do the scanning.  We parse the output here because it seems that
# f-prot is not returning the error code?
my $fp_message = `$fprot $tmpdir`;

# Also check for nasty extensions (you will need to zip/pw protect to bypass these)
my %cf_bad = ();
# Read bad array
open( IN, "$FindBin::Bin/bad.ext" );
while( <IN> ) {
	next if /^.*#/; # skip comments
	s/^\s+//;
	s/\s+$//; # clean spaces
	$cf_bad{ $_ } = 1;
}
close( IN );

my $cf_isbad = "";
for( <$tmpdir/*> ) {
       my( $cf_ext ) = m/\.([^.]+)$/;
       if( $cf_bad{ lc( $cf_ext ) } ) {
               $cf_isbad .= "$_ is not an allowed file type\n";
       }
}

# Scan inside zip files
for( <$tmpdir/*.[Zz][Ii][Pp]> ) {
       my $z = Archive::Zip -> new( $_ );
       for( $z -> memberNames() ) {
               my( $ext ) = m/\.([^.]+)$/;
               if( $cf_bad{ lc( $ext ) } ) {
                       $cf_isbad .= "$_ is not an allowed file type (inside zip)\n";
               }
       }
}

if( $cf_isbad ) {
       $fp_message .= "-" x 20;
       $fp_message .= "\n$cf_isbad\n";
}

# From the changes file:
#Note 2: We received several requests for a list of return codes for the
#        Linux version, and here they are:
#
#        0  Normal exit.  Nothing found, nothing done.
#        1  Unrecoverable error (for example, missing SIGN.DEF).
#        2  Selftest failed (program has been modified).
#        3  At least one virus-infected object was found.
#        4  <not used>
#        5  Abnormal termination (scanning did not finish).
#        6  At least one virus was removed.
#        7  Error, out of memory (should never happen, but well...)
#        8  Something suspicious was found, but no recognized virus.
#
# So we'll error on 3 and 6.  Can't really error on other codes...
# we need to let the mail pass through.  Maybe add a suspicious flag?
my $exit_status = ( $? >> 8 );

# Check, are we OK?
if( ( $exit_status == 3 ) ||
	( $exit_status == 6 ) ||
	( $cf_isbad ) ) {
	# BAD!
	$virus_found = 1;

	if( $quarantine ) {
		# Make a copy of this message
		my $qname;

		open( SAVEMSG, ">$tmpdir/fullmessage.txt" );
		$en -> print( \*SAVEMSG );
		close( SAVEMSG );

		open( SAVELOG, ">$tmpdir/virus.log" );
		print SAVELOG $fp_message;
		close( SAVELOG );

		$qname = "$scan_path/quarantine/";
		$qname .= scalar( time() );
		$qname .= ".$$";
		mkdir $qname, 0755;
		logmsg( "Making quarantine: $qname" );
		system( "cp -a $tmpdir/* $qname" );
	}

	$en -> purge();

	my $message;
	
	if( $opts{ m } )
	{
		open( IN, $opts{ m } );
		$message .= $_ while <IN>;
		close( IN );
	}
	
	if( ! $message )
	{
		$message = qq{
A mail message with infected attachments has been detected.  If you 
know the person who allegedly sent this email you might want to contact 
them and let them know they may have a virus.  If you have any questions
about this email, or perhaps want to try and salvage the email if you
think it may have been valuable, please contact your mail administrator.
A copy of the scan log is below.

	Thanks,
   		VScan Mail Scanner

-------------------
};
	}

	$en = MIME::Entity -> build( 
		From => $administrator,
		To => $to,
		Subject => "INFECTED EMAIL: from $from regarding $subject",
		Data => [ "$message\n$fp_message" ]
	);

	# Send off the admin mail
	mail_admin( 'VIRUS ALERT', qq{
Summary:

     To: $to
   From: $from
Subject: $subject

Environment:
$edump

The recipient has been notified.  The virus log:

$fp_message
} ) unless $opts{ S };
}

if( $always_attach_log )
{
	# Attach the virus log
	$en -> attach(
		Path => $tmpdir . "_log.txt",
		Type => 'text/plain'
	);
}

# At this point, any bad parts will be modified, so reconstruct the message
$en -> head() -> replace( 'x-security', "VScan v$VER - http://hoopajoo.net" );
$en -> head() -> replace( 'x-virus-found',  $virus_found ? 'yes' : 'no' );

$en -> sync_headers( Length => 'COMPUTE' );
print $first_line;
$en -> print( \*STDOUT );

# Purge
$en -> purge();

# Clean up log/file
logmsg( "Deleting $tmpdir" );
system( "/bin/rm", "-rf", $tmpdir ) if -d $tmpdir;

sub logmsg
{
	print STDERR shift() . "\n";
}

# Can attach up to 1 file.
sub mail_admin
{
	my $subject = shift;
	my $message = shift;
	my $attach = shift;

	my $mailer = Mail::Sender -> new( {
		smtp => 'localhost',
		from => $administrator
	} );

	if( $attach )
	{
		$mailer -> MailFile( {
			to => $administrator,
			subject => $subject,
			msg => $message,
			file => $attach,
			ctype => 'text/plain'
		} );
	}else
	{
		$mailer -> MailMsg( {
			to => $administrator,
			subject => $subject,
			msg => $message
		} );
	}
}

